#define zerodetect 2        //детектор нуля
#define counterpwr 3
#define triac A3        //тиристор
#define selectIU 4
#define counterIU 5
#define POWER_TO_REAL(x) (((x) * 10) / 10)
#define U_TO_REAL(x) (((x) * 10) / 10)


#include <LiquidCrystal.h>
LiquidCrystal lcd(12,11,10,9,7,6);


#include <Wire.h>
#define SLAVE_ADDRESS       0x04        // Адрес шины i2c для связи с внешним миром. Slave.

int power_ticks;     // счетчик импульсов по пину counterpwr
long int power_set;       // заданная мощность (x1,8)
long int power_real;      // измеренная мощность (x1,8)
long int energy;
#define power_step   180 // 100 ватт
#define power_min   0
#define power_max   17820 // 9990 ватт

int ui_real;

byte timer2_16bit;   // служебный счетчик для таймер2

long int triac_time;       // время через которое должен открыться тиристор      
byte triac_time_h;    // старший байт
#define triac_time_max 2440 // в теории 2500 предел. отступим на 60 единиц (240мкс) чтоб детектор нуля успевал. = 4,3% от возможного
#define triac_time_min 60 // отступим немного. максимальная мощность будет более 95% 
#define triac_step_max 100

//---------------------------------------------------------------------------------------
// Константы - состояние кнопок
#define NO_ACT              0
#define PLUS_CLICK          0b00000001  // Короткое нажатие кнопки "+"
#define MINUS_CLICK         0b00000010  // Короткое нажатие кнопки "-"
#define PLUS_PRESS          0b00000100  // Удержание кнопки "+" в нажатом состоянии
#define MINUS_PRESS         0b00001000  // Удержание кнопки "-" в нажатом состоянии

#define mButPin  A1             // Пин, к которому подключена кнопка "-"
#define pButPin  A0            // Пин, к которому подключена кнопка "+"

#define dtBounce 10 // Задержка (мс) для устранения дребезга контактов кнопок
#define dtPress 300 // Задержка, после которой считается, что кнопка удерживается
static uint8_t bos = NO_ACT; // Предыдущее состояние кнопок
static unsigned long tpr; // Момент нажатия кнопки




byte need_update;
int time_ticks; // тики времени

void receiveData(int byteCount); // callback для принимаемых по i2c байтов
void sendData(); // callback для передаваемых по i2c байтов
void updateDisplay(); 

uint8_t GetKey(void) {
  uint8_t ret, bcs = ((!digitalRead(mButPin) & 0x01) << 1) | (!digitalRead(pButPin) & 0x01); // Текущее состояние кнопок
  if(!bos && bcs) { // Произошло нажатие кнопки
    tpr = millis(); bos = bcs; // Фиксируем время и состояние кнопок
    return NO_ACT;
  }
  if(bos && !bcs) { // Произошло либо отжатие или дребезг кнопки
    if((millis() - tpr) < dtBounce) return NO_ACT; // Дребезг, однако...
    ret = bos; bos = bcs; return ret; // Отжатие...
  }
  if(bos && bcs) { // Возможно, удерживание кнопки
    if((millis() - tpr) < dtPress) return NO_ACT; // Но пока еще не ясно...
    // И все-таки это удержание кнопки в нажатом состоянии
    ret = bos << 2; tpr = millis(); return ret;
  }
  return NO_ACT;
}

void zero_cross_int()
{
  TCNT2 = 0;
  timer2_16bit = 0;
}

void power_count()
{
  power_ticks++;
}

ISR (TIMER0_COMPA_vect)  // Обработчик прерывания от таймера TIMER0
// тикать будет с частотой 976.5625 гц
// будем синкаться от него, чтоб не было "ноу зеро детект"
{
  time_ticks++;
  if (time_ticks >= 488) time_ticks = 0; // пол секунды с погрешностью 0.05%
  if (time_ticks == 0) need_update = 1;
  
  uint8_t ks = GetKey();
  long int power_tmp = power_set;
  switch(ks) {
    case PLUS_CLICK:  if(power_tmp >= power_max) power_tmp = power_max;
                      else power_tmp += 18;
                      break;
    case MINUS_CLICK: if(power_tmp <= power_min) power_tmp = power_min;
                      else power_tmp -= 18;
                      break;
    case PLUS_PRESS:  if(power_tmp >= power_max - power_step) power_tmp = power_max;
                      else power_tmp += power_step;
                      break;
    case MINUS_PRESS: if(power_tmp <= power_min + power_step) power_tmp = power_min;
                      else power_tmp -= power_step;
                      break;
  }
  if(power_set != power_tmp)
  {
    noInterrupts();
    power_set = power_tmp;
    interrupts();
    updateDisplay();
  }
  
}

ISR (TIMER2_COMPA_vect)  // Обработчик прерывания от таймера TIMER2
{
  // вызывается с частотой 976 гц. один тик = 1мс
  timer2_16bit++;
  if ((timer2_16bit == triac_time_h) & (power_set > 0))
  {
    PORTC |=  (1 << PORTC3);
  };
}
ISR (TIMER2_COMPB_vect)  // Обработчик прерывания от таймера TIMER2
{
  // вызывается с частотой 976 гц. один тик = 1мс
  PORTC &=  ~(1 << PORTC3);
}

//---------------------------------------------------------------------------------------
void receiveData(int byteCount) { // callback для принимаемых по i2c байтов. Через i2c можно задать мощность не более 2550 Вт
  while (Wire.available()) {
    byte tmp = Wire.read();
    power_set = (int)(tmp)*18;
  }
}
//---------------------------------------------------------------------------------------
void sendData() { // callback для передаваемых по i2c байтов
  byte tmp = (byte)(power_real / 18);
  if((power_real / 18) > 255) tmp = 255;
  Wire.write(tmp);
}
//---------------------------------------------------------------------------------------


void setup() {

  Wire.begin(SLAVE_ADDRESS); // Инициализация библиотеки для шины i2c
  Wire.onReceive(receiveData);
  Wire.onRequest(sendData);


  pinMode(zerodetect, INPUT);     //детектор нуля
  pinMode(counterpwr, INPUT);     //счетчик мощности
  pinMode(counterIU, INPUT);     //счетчик напряжения или тока
  pinMode(triac, OUTPUT);      //тиристор
  pinMode(selectIU, OUTPUT);      //выбор что замерять, напряжение или ток
  digitalWrite(selectIU, 0);
  
  pinMode(mButPin, INPUT_PULLUP); // Инициализация пинов, к которым подключены кнопки
  pinMode(pButPin, INPUT_PULLUP);

  pinMode(13, OUTPUT);      //блинк

  timer2_16bit = 0;
  power_ticks = 0;
  power_set = 1800;
  triac_time = 2399;
  triac_time_h = (byte)(triac_time / 256);
  energy = 0; // копим энергию 


  // таймер0. на него завязаны delay и millis
  // добавляем в него тики по совпадению. тикать будет с частотой 976 гц. сдвигаем его на пол периода от переполнения.
  OCR0A = 127;
  TIMSK0 |= (1 << OCIE0A); // Разрешить прерывание по совпадению

  // таймер1 переключаем на внешний захват по T1. будем им напряжение или ток замерять
  TCCR1A = 0x00;  //
  TCCR1B = (1 << CS12) | (1 << CS11) | (0 << CS10); // Тактирование от T1. External clock source on T1 pin. Clock on falling edge

  // таймер2 используем для замеров отрытия триака. частоту оставим по умолчанию 125кгц
  // недостающие 8 бит проэмулируем софтово
  TCCR2A = 0; 
  TIMSK2 |= (1 << OCIE2A); // Разрешить прерывание по совпадению
  TIMSK2 |= (1 << OCIE2B); // Разрешить прерывание по совпадению
  OCR2A = 0;


  attachInterrupt(digitalPinToInterrupt(counterpwr), power_count, RISING);//вызов прерывания счетчика мощности
  attachInterrupt(digitalPinToInterrupt(zerodetect), zero_cross_int, RISING);//вызов прерывания при детектировании нуля

  lcd.begin(8, 2);

  //Serial.begin(57600);

}

byte i = 0;

void updateDisplay()
{
  
    lcd.setCursor(0, 0);
    lcd.print("P");
    long int p_norm;
    p_norm = (power_real * 100) / 1800;
    if(p_norm < 10) lcd.print(" ");
    if(p_norm < 100) lcd.print(" ");    
    lcd.print(p_norm);
    lcd.print("/");
    p_norm = (power_set * 100) / 1800;
    if(p_norm < 10) lcd.print(" ");
    if(p_norm < 100) lcd.print(" ");    
    lcd.print(p_norm);
    

    lcd.setCursor(0, 1);
    if ((i/4) % 2) 
    { 
      int triac_time_tmp = 100 - (triac_time / 25);
      //triac_time_tmp = triac_time;
      if(triac_time_tmp < 10) lcd.print(" ");    
      lcd.print(triac_time_tmp);
      lcd.print("% U");
      long int u_norm = (long int)ui_real * 1000 / 4589;
      if(u_norm < 10) lcd.print(" ");
      if(u_norm < 100) lcd.print(" ");    
      lcd.print(u_norm);
    }
    else
    {
      long int energywh;
      energywh = energy / 1305; 
      lcd.print(energywh);
      lcd.print("Wh      ");
    }
}

void loop() {
  // put your main code here, to run repeatedly:
  if (need_update)
  {
    noInterrupts();
    power_real = power_ticks; 
    power_ticks = 0;
    ui_real = TCNT1;
    TCNT1 = 0;
    interrupts();
    energy += power_real;
    power_real *= 10;
    
    long int triac_step;
    if (power_real > 10) 
      triac_step = (2500 - ((2500 - triac_time) * power_set )/ (power_real) - triac_time)/2;
      else
      triac_step = -20;


    if (triac_step > triac_step_max) {triac_step = triac_step_max;};
    if (triac_step < -triac_step_max) {triac_step = -triac_step_max;};
    triac_time += triac_step; 

   
    
    if (triac_time > triac_time_max) triac_time = triac_time_max;
    if (triac_time < triac_time_min) triac_time = triac_time_min;

    triac_time_h = (byte)(triac_time / 256);
    OCR2A = (byte)(triac_time % 256);
    OCR2B = (byte)(triac_time % 256) + 3; // 250кгц, один тик 4мкс, делаем задержку 3 тика = 12мкс. этого достаточно для открытия моськи
    
    updateDisplay();


    i++;
    digitalWrite(13, i % 2);
    need_update = 0;
  };

}
